@@ -1,5 +1,4 @@
 #include "TeleBoy.h"
-#include "Cache.h"
 #include "md5.h"
 #include "Utils.h"
 #ifdef TARGET_WINDOWS
@@ -27,84 +26,7 @@ using namespace std;
 using namespace rapidjson;
 
 static const string apiUrl = "https://tv.api.teleboy.ch";
-static const string apiDeviceType = "desktop";
-static const string apiVersion = "2.0";
-const char data_file[] = "special://profile/addon_data/pvr.teleboy/data.json";
 std::mutex TeleBoy::sendEpgToKodiMutex;
-static const std::string user_agent = std::string("Kodi/")
-    + std::string(STR(KODI_VERSION)) + std::string(" pvr.teleboy/")
-    + std::string(STR(TELEBOY_VERSION)) + std::string(" (Kodi PVR addon)");
-
-
-std::string TeleBoy::HttpGetCached(Curl &curl, const std::string& url, time_t cacheDuration)
-{
-
-  std::string content;
-  std::string cacheKey = md5(url);
-  if (!Cache::Read(cacheKey, content))
-  {
-    content = HttpGet(curl, url);
-    if (!content.empty())
-    {
-      time_t validUntil;
-      time(&validUntil);
-      validUntil += cacheDuration;
-      Cache::Write(cacheKey, content, validUntil);
-    }
-  }
-  return content;
-}
-
-string TeleBoy::HttpGet(Curl &curl, string url)
-{
-  return HttpRequest(curl, "GET", url, "");
-}
-
-string TeleBoy::HttpDelete(Curl &curl, string url)
-{
-  return HttpRequest(curl, "DELETE", url, "");
-}
-
-string TeleBoy::HttpPost(Curl &curl, string url, string postData)
-{
-  return HttpRequest(curl, "POST", url, postData);
-}
-
-string TeleBoy::HttpRequest(Curl &curl, string action, string url,
-    string postData)
-{
-  curl.AddHeader("User-Agent", user_agent);
-  int statusCode;
-  kodi::Log(ADDON_LOG_DEBUG, "Http-Request: %s %s.", action.c_str(), url.c_str());
-  string content;
-  if (action.compare("POST") == 0)
-  {
-    content = curl.Post(url, postData, statusCode);
-  }
-  else if (action.compare("DELETE") == 0)
-  {
-    content = curl.Delete(url, statusCode);
-  }
-  else
-  {
-    content = curl.Get(url, statusCode);
-  }
-  string cinergys = curl.GetCookie("cinergy_s");
-  if (!cinergys.empty() && cinergys != cinergySCookies && cinergys != "deleted")
-  {
-    cinergySCookies = cinergys;
-    WriteDataJson();
-  }
-  return content;
-}
-
-void TeleBoy::ApiSetHeader(Curl &curl)
-{
-  curl.AddHeader("x-teleboy-apikey", apiKey);
-  curl.AddHeader("x-teleboy-device-type", apiDeviceType);
-  curl.AddHeader("x-teleboy-session", cinergySCookies);
-  curl.AddHeader("x-teleboy-version", apiVersion);
-}
 
 bool TeleBoy::ApiGetResult(string content, Document &doc)
 {
@@ -115,52 +37,55 @@ bool TeleBoy::ApiGetResult(string content, Document &doc)
     {
       return true;
     }
+    if (doc["error_code"].GetInt() == 10403) {
+      kodi::Log(ADDON_LOG_WARNING, "Got error_code 10403. Reset session.");
+      m_session->Reset();
+    }
   }
   return false;
 }
 
 bool TeleBoy::ApiGet(string url, Document &doc, time_t timeout)
 {
-  Curl curl;
-  ApiSetHeader(curl);
   string content;
+  int statusCode;
+  if (!m_session->IsConnected()) {
+    return false;
+  }
   if (timeout > 0) {
-    content = HttpGetCached(curl, apiUrl + url, timeout);
+    content = m_httpClient->HttpGetCached(apiUrl + url, timeout, statusCode);
   } else {
-    content = HttpGet(curl, apiUrl + url);
+    content = m_httpClient->HttpGet(apiUrl + url, statusCode);
   }
-  curl.ResetHeaders();
   return ApiGetResult(content, doc);
 }
 
 bool TeleBoy::ApiPost(string url, string postData, Document &doc)
 {
-  Curl curl;
-  ApiSetHeader(curl);
-  if (!postData.empty())
-  {
-    curl.AddHeader("Content-Type", "application/json");
+  int statusCode;
+  if (!m_session->IsConnected()) {
+    return false;
   }
-  string content = HttpPost(curl, apiUrl + url, postData);
-  curl.ResetHeaders();
+  string content = m_httpClient->HttpPost(apiUrl + url, postData, statusCode);
   return ApiGetResult(content, doc);
 }
 
 bool TeleBoy::ApiDelete(string url, Document &doc)
 {
-  Curl curl;
-  ApiSetHeader(curl);
-  string content = HttpDelete(curl, apiUrl + url);
-  curl.ResetHeaders();
+  int statusCode;
+  if (!m_session->IsConnected()) {
+    return false;
+  }
+  string content = m_httpClient->HttpDelete(apiUrl + url, statusCode);
   return ApiGetResult(content, doc);
 }
 
-TeleBoy::TeleBoy() :
-    teleboyUsername(""), teleboyPassword(""), maxRecallSeconds(60 * 60 * 24 * 7), cinergySCookies(
-        ""), isPlusMember(false), isComfortMember(false)
+TeleBoy::TeleBoy()
 {
-  kodi::Log(ADDON_LOG_INFO, "Using useragent: %s", user_agent.c_str());
-  ReadDataJson();
+  m_parameterDB = new ParameterDB(UserPath());
+  m_httpClient = new HttpClient(m_parameterDB);
+  m_session = new Session(m_httpClient, this);
+  m_httpClient->SetStatusCodeHandler(m_session);
 }
 
 TeleBoy::~TeleBoy()
@@ -169,172 +94,41 @@ TeleBoy::~TeleBoy()
   {
     delete updateThread;
   }
+  delete m_httpClient;
+  delete m_parameterDB;
 }
 
 ADDON_STATUS TeleBoy::Create()
 {
   kodi::Log(ADDON_LOG_DEBUG, "%s - Creating the PVR Teleboy add-on", __FUNCTION__);
-
-  favoritesOnly = kodi::GetSettingBoolean("favoritesonly");
-  enableDolby = kodi::GetSettingBoolean("enableDolby");
-  teleboyUsername = kodi::GetSettingString("username");
-  teleboyPassword = kodi::GetSettingString("password");
-
-  if (teleboyUsername.empty() || teleboyPassword.empty())
-  {
-    kodi::Log(ADDON_LOG_INFO, "Username or password not set.");
-    kodi::QueueNotification(QUEUE_WARNING, "", kodi::GetLocalizedString(30100));
-    return ADDON_STATUS_NEED_SETTINGS;
-  }
-
-  kodi::Log(ADDON_LOG_DEBUG, "Login Teleboy");
-  if (Login(teleboyUsername, teleboyPassword))
-  {
-    kodi::Log(ADDON_LOG_DEBUG, "Login done");
-  }
-  else
-  {
-    kodi::Log(ADDON_LOG_ERROR, "Login failed");
-    kodi::QueueNotification(QUEUE_ERROR, "", kodi::GetLocalizedString(30101));
-    return ADDON_STATUS_NEED_SETTINGS;
-  }
-
-  return ADDON_STATUS_OK;
+  return m_session->Start();
 }
 
 ADDON_STATUS TeleBoy::SetSetting(const std::string& settingName, const kodi::CSettingValue& settingValue)
 {
-  if (settingName == "username")
-  {
-    string username = settingValue.GetString();
-    if (username != teleboyUsername)
-    {
-      teleboyUsername = username;
-      return ADDON_STATUS_NEED_RESTART;
-    }
-  }
+  return m_session->SetSetting(settingName, settingValue);
+}
 
-  if (settingName == "password")
-  {
-    string password = settingValue.GetString();
-    if (password != teleboyPassword)
-    {
-      teleboyPassword = password;
-      return ADDON_STATUS_NEED_RESTART;
-    }
-  }
+void TeleBoy::UpdateConnectionState(const std::string& connectionString, PVR_CONNECTION_STATE newState, const std::string& message) {
+  kodi::addon::CInstancePVRClient::ConnectionStateChange(connectionString, newState, message);
+}
 
-  if (settingName == "favoritesonly")
-  {
-    bool favOnly = settingValue.GetBoolean();
-    if (favOnly != favoritesOnly)
-    {
-      favoritesOnly = favOnly;
-      return ADDON_STATUS_NEED_RESTART;
-    }
+ADDON_STATUS TeleBoy::GetStatus() {
+  if (!m_session->IsConnected()) {
+    return ADDON_STATUS_LOST_CONNECTION;
   }
-
   return ADDON_STATUS_OK;
 }
 
-bool TeleBoy::Login(string u, string p)
+void TeleBoy::SessionInitialized()
 {
-  string tbUrl = "https://www.teleboy.ch";
-  Curl curl;
-  if (!cinergySCookies.empty())
+  while (updateThreads.size() < 3)
   {
-    curl.AddOption("cookie", "cinergy_s=" + cinergySCookies);
-  }
-  string result = HttpGet(curl, tbUrl + "/live");
-  bool isAuthenticated = result.find("setIsAuthenticated(true") != std::string::npos;
-  curl.AddHeader("redirect-limit", "0");
-
-  if (!isAuthenticated) {
-    kodi::Log(ADDON_LOG_INFO, "Not yet authenticated. Try to login.");
-    HttpGet(curl, tbUrl + "/login");
-    string location = curl.GetLocation();
-    if (location.find("t.teleboy.ch") != string::npos)
-    {
-      kodi::Log(ADDON_LOG_INFO, "Using t.teleboy.ch.");
-      tbUrl = "https://t.teleboy.ch";
-      HttpGet(curl, tbUrl + "/login");
-    }
-    curl.AddHeader("Referer", tbUrl + "/login");
-    if (!cinergySCookies.empty())
-    {
-      curl.AddOption("cookie", "cinergy_s=" + cinergySCookies);
-    }
-    result = HttpPost(curl, tbUrl + "/login_check",
-        "login=" + Utils::UrlEncode(u) + "&password=" + Utils::UrlEncode(p)
-            + "&keep_login=1");
-    curl.ResetHeaders();
-    curl.AddHeader("redirect-limit", "5");
-    curl.AddHeader("Referer", tbUrl + "/login");
-    if (!cinergySCookies.empty())
-    {
-      curl.AddOption("cookie", "welcomead=1; cinergy_s=" + cinergySCookies);
-    }
-    result = HttpGet(curl, tbUrl);
-    curl.ResetHeaders();
-    if (result.empty())
-    {
-      kodi::Log(ADDON_LOG_ERROR, "Failed to login.");
-      return false;
-    }
-  } else {
-    kodi::Log(ADDON_LOG_INFO, "Already authenticated.");
-  }
-
-  size_t pos = result.find("tvapiKey:");
-  size_t pos1 = result.find("'", pos) + 1;
-  if (pos == std::string::npos || pos1 > pos + 50)
-  {
-    kodi::Log(ADDON_LOG_ERROR, "No api key found.");
-    return false;
-  }
-  size_t endPos = result.find("'", pos1);
-  if (endPos - pos1 > 65 || endPos <= pos)
-  {
-    kodi::Log(ADDON_LOG_DEBUG, "Got HTML body: %s", result.c_str());
-    kodi::Log(ADDON_LOG_ERROR, "Received api key is invalid.");
-    return false;
-  }
-  apiKey = result.substr(pos1, endPos - pos1);
-
-  pos = result.find("setId(");
-  if (pos == std::string::npos)
-  {
-    kodi::Log(ADDON_LOG_ERROR, "No user settings found.");
-    return false;
-  }
-  pos += 6;
-  endPos = result.find(")", pos);
-  if (endPos - pos > 15 || endPos <= pos)
-  {
-    kodi::Log(ADDON_LOG_DEBUG, "Got HTML body: %s", result.c_str());
-    kodi::Log(ADDON_LOG_ERROR, "Received userId is invalid.");
-    return false;
-  }
-  userId = result.substr(pos, endPos - pos);
-
-  isPlusMember = result.find("setIsPlusMember(1", endPos) != std::string::npos;
-  isComfortMember = result.find("setIsComfortMember(1", endPos)
-      != std::string::npos;
-  if (!isPlusMember) {
-    kodi::Log(ADDON_LOG_INFO, "Free accounts are not supported.", userId.c_str());
-    kodi::QueueNotification(QUEUE_ERROR, "", kodi::GetLocalizedString(30102));
-    return false;
-  }
-  kodi::Log(ADDON_LOG_DEBUG, "Got userId: %s.", userId.c_str());
-
-  for (int i = 0; i < 3; ++i)
-  {
-    updateThreads.emplace_back(new UpdateThread(i, *this));
+    updateThreads.emplace_back(new UpdateThread(updateThreads.size(), *this, *m_session));
   }
 
   LoadChannels();
   LoadGenres();
-  return true;
 }
 
 PVR_ERROR TeleBoy::GetCapabilities(kodi::addon::PVRCapabilities& capabilities)
@@ -370,7 +164,7 @@ PVR_ERROR TeleBoy::GetBackendVersion(std::string& version)
 
 PVR_ERROR TeleBoy::GetConnectionString(std::string& connection)
 {
-  connection = "connected";
+  connection = m_session->IsConnected() ? "connected" : "not connected";
   return PVR_ERROR_NO_ERROR;
 }
 
@@ -435,7 +229,7 @@ bool TeleBoy::LoadChannels()
     channelsById[channel.id] = channel;
   }
 
-  if (!ApiGet("/users/" + userId + "/stations", json, 3600))
+  if (!ApiGet("/users/" + m_session->GetUserId() + "/stations", json, 3600))
   {
     kodi::Log(ADDON_LOG_ERROR, "Error loading sorted channels.");
     return false;
@@ -455,7 +249,11 @@ bool TeleBoy::LoadChannels()
 
 PVR_ERROR TeleBoy::GetChannelsAmount(int& amount)
 {
-  if (favoritesOnly)
+  if (!m_session->IsConnected()) {
+    return PVR_ERROR_SERVER_ERROR;
+  }
+
+  if (m_session->GetFavoritesOnly())
   {
     amount = sortedChannels.size();
   }
@@ -468,13 +266,17 @@ PVR_ERROR TeleBoy::GetChannelsAmount(int& amount)
 
 PVR_ERROR TeleBoy::GetChannels(bool radio, kodi::addon::PVRChannelsResultSet& results)
 {
+  if (!m_session->IsConnected()) {
+    return PVR_ERROR_SERVER_ERROR;
+  }
+
   int channelNum = 0;
   for (int const &cid : sortedChannels)
   {
     channelNum++;
     TransferChannel(results, channelsById[cid], channelNum);
   }
-  if (!favoritesOnly)
+  if (!m_session->GetFavoritesOnly())
   {
     for (auto const &item : channelsById)
     {
@@ -504,39 +306,55 @@ void TeleBoy::TransferChannel(kodi::addon::PVRChannelsResultSet& results, TeleBo
   results.Add(kodiChannel);
 }
 
-void TeleBoy::SetStreamProperties(std::vector<kodi::addon::PVRStreamProperty>& properties, const std::string& url, bool realtime)
+PVR_ERROR TeleBoy::SetStreamProperties(std::vector<kodi::addon::PVRStreamProperty>& properties, const Value& stream, bool realtime)
 {
+  string url = GetStringOrEmpty(stream, "url");
+  kodi::Log(ADDON_LOG_INFO, "Play URL: %s.", url.c_str());
+  url = FollowRedirect(url);
+
+  if (url.empty())
+  {
+    return PVR_ERROR_FAILED;
+  }
+
   properties.emplace_back(PVR_STREAM_PROPERTY_STREAMURL, url);
   properties.emplace_back(PVR_STREAM_PROPERTY_INPUTSTREAM, "inputstream.adaptive");
   properties.emplace_back("inputstream.adaptive.manifest_type", "mpd");
   properties.emplace_back("inputstream.adaptive.manifest_update_parameter", "full");
   properties.emplace_back(PVR_STREAM_PROPERTY_MIMETYPE, "application/xml+dash");
   properties.emplace_back(PVR_STREAM_PROPERTY_ISREALTIMESTREAM, realtime ? "true" : "false");
+      
+  if (stream.HasMember("drm")) {
+    string drmType = GetStringOrEmpty(stream["drm"], "type");
+    if (drmType == "widevine") {
+      string licenseUrl = GetStringOrEmpty(stream["drm"], "license_url");
+      properties.emplace_back("inputstream.adaptive.license_key", licenseUrl + "||A{SSM}|");
+      properties.emplace_back("inputstream.adaptive.license_type", "com.widevine.alpha"); 
+    } else {
+      kodi::Log(ADDON_LOG_ERROR, "Unsupported drm type: %s.", drmType.c_str());
+    }      
+  }
+  return PVR_ERROR_NO_ERROR;
 }
 
 PVR_ERROR TeleBoy::GetChannelStreamProperties(const kodi::addon::PVRChannel& channel, std::vector<kodi::addon::PVRStreamProperty>& properties)
 {
-  PVR_ERROR ret = PVR_ERROR_FAILED;
+  if (!m_session->IsConnected()) {
+    return PVR_ERROR_SERVER_ERROR;
+  }
 
   Document json;
   if (!ApiGet(
-      "/users/" + userId + "/stream/live/" + to_string(channel.GetUniqueId())
+      "/users/" + m_session->GetUserId() + "/stream/live/" + to_string(channel.GetUniqueId())
           + "?expand=primary_image,flags&https=1" + GetStreamParameters(), json, 0))
   {
     kodi::Log(ADDON_LOG_ERROR, "Error getting live stream url for channel %i.",
         channel.GetUniqueId());
-    return ret;
+    return PVR_ERROR_FAILED;
   }
-  string url = GetStringOrEmpty(json["data"]["stream"], "url");
-  kodi::Log(ADDON_LOG_INFO, "Play URL: %s.", url.c_str());
-  url = FollowRedirect(url);
+  const Value& stream = json["data"]["stream"];
+  return SetStreamProperties(properties, stream, true);
 
-  if (!url.empty())
-  {
-    SetStreamProperties(properties, url, true);
-    ret = PVR_ERROR_NO_ERROR;
-  }
-  return ret;
 }
 
 string TeleBoy::FollowRedirect(string url)
@@ -575,7 +393,7 @@ void TeleBoy::GetEPGForChannelAsync(int uniqueChannelId, time_t iStart,
   {
     Document json;
     if (!ApiGet(
-        "/users/" + userId + "/broadcasts?begin=" + FormatDate(iStart)
+        "/users/" + m_session->GetUserId() + "/broadcasts?begin=" + FormatDate(iStart)
             + "+00:00:00&end=" + FormatDate(iEnd + 60 * 60 * 24) + "+00:00:00&expand=logos&limit=500&skip="
             + to_string(sum) + "&sort=station&station="
             + to_string(uniqueChannelId), json, 60*60*24))
@@ -658,8 +476,11 @@ PVR_ERROR TeleBoy::GetRecordingsAmount(bool deleted, int& amount)
 
 PVR_ERROR TeleBoy::DeleteRecording(const kodi::addon::PVRRecording& recording)
 {
+  if (!m_session->IsConnected()) {
+    return PVR_ERROR_SERVER_ERROR;
+  }
   Document doc;
-  if (!ApiDelete("/users/" + userId + "/recordings/" + recording.GetRecordingId(), doc))
+  if (!ApiDelete("/users/" + m_session->GetUserId() + "/recordings/" + recording.GetRecordingId(), doc))
   {
     kodi::Log(ADDON_LOG_ERROR, "Error deleting recording %s.", recording.GetRecordingId().c_str());
     return PVR_ERROR_SERVER_ERROR;
@@ -669,6 +490,10 @@ PVR_ERROR TeleBoy::DeleteRecording(const kodi::addon::PVRRecording& recording)
 
 PVR_ERROR TeleBoy::GetRecordings(bool deleted, kodi::addon::PVRRecordingsResultSet& results)
 {
+  if (!m_session->IsConnected()) {
+    return PVR_ERROR_SERVER_ERROR;
+  }
+
   int totals = -1;
   int sum = 0;
   string type = "ready";
@@ -676,7 +501,7 @@ PVR_ERROR TeleBoy::GetRecordings(bool deleted, kodi::addon::PVRRecordingsResultS
   {
     Document json;
     if (!ApiGet(
-        "/users/" + userId + "/recordings/" + type
+        "/users/" + m_session->GetUserId() + "/recordings/" + type
             + "?desc=1&expand=flags,logos&limit=100&skip=" + to_string(sum) + "&sort=date", json, 10))
     {
       kodi::Log(ADDON_LOG_ERROR, "Error getting recordings of type %s.",
@@ -693,12 +518,10 @@ PVR_ERROR TeleBoy::GetRecordings(bool deleted, kodi::addon::PVRRecordingsResultS
 
       kodi::addon::PVRRecording tag;
 
-      tag.SetSeriesNumber(PVR_RECORDING_INVALID_SERIES_EPISODE);
-      tag.SetEpisodeNumber(PVR_RECORDING_INVALID_SERIES_EPISODE);
       tag.SetIsDeleted(false);
       tag.SetRecordingId(to_string(item["id"].GetInt()));
       tag.SetTitle(GetStringOrEmpty(item, "title"));
-      tag.SetEpisodeName(GetStringOrEmpty(item, "subtitle"));
+      tag.SetEpisodeName(GetStringOrEmpty(item, "subtitle"));      
       tag.SetPlot(GetStringOrEmpty(item, "description"));
       tag.SetPlotOutline(GetStringOrEmpty(item, "short_description"));
       tag.SetChannelUid(item["station_id"].GetInt());
@@ -708,6 +531,13 @@ PVR_ERROR TeleBoy::GetRecordings(bool deleted, kodi::addon::PVRRecordingsResultS
       time_t endTime = Utils::StringToTime(GetStringOrEmpty(item, "end"));
       tag.SetDuration(endTime - tag.GetRecordingTime());
       tag.SetEPGEventId(item["id"].GetInt());
+      if (item.HasMember("serie_season")) {
+        tag.SetSeriesNumber(item["serie_season"].GetInt());
+        tag.SetDirectory(tag.GetTitle());
+      }
+      if (item.HasMember("serie_episode")) {
+        tag.SetEpisodeNumber(item["serie_episode"].GetInt());
+      }
       if (item.HasMember("genre_id")) {
         int genreId = item["genre_id"].GetInt();
         TeleboyGenre genre = genresById[genreId];
@@ -730,24 +560,21 @@ PVR_ERROR TeleBoy::GetRecordings(bool deleted, kodi::addon::PVRRecordingsResultS
 
 PVR_ERROR TeleBoy::GetRecordingStreamProperties(const kodi::addon::PVRRecording& recording, std::vector<kodi::addon::PVRStreamProperty>& properties)
 {
+  if (!m_session->IsConnected()) {
+    return PVR_ERROR_SERVER_ERROR;
+  }
+
   PVR_ERROR ret = PVR_ERROR_FAILED;
 
   Document json;
-  if (!ApiGet("/users/" + userId + "/stream/" + recording.GetRecordingId() + "?" + GetStreamParameters(), json, 0))
+  if (!ApiGet("/users/" + m_session->GetUserId() + "/stream/" + recording.GetRecordingId() + "?" + GetStreamParameters(), json, 0))
   {
     kodi::Log(ADDON_LOG_ERROR, "Could not get URL for recording: %s.",
         recording.GetRecordingId().c_str());
     return ret;
   }
-  string url = GetStringOrEmpty(json["data"]["stream"], "url");
-  url = FollowRedirect(url);
-
-  if (!url.empty())
-  {
-    SetStreamProperties(properties, url, false);
-    ret = PVR_ERROR_NO_ERROR;
-  }
-  return ret;
+  const Value& stream = json["data"]["stream"];
+  return SetStreamProperties(properties, stream, false);
 }
 
 PVR_ERROR TeleBoy::GetRecordingEdl(const kodi::addon::PVRRecording& recording, std::vector<kodi::addon::PVREDLEntry>& edl)
@@ -775,6 +602,10 @@ PVR_ERROR TeleBoy::GetTimersAmount(int& amount)
 
 PVR_ERROR TeleBoy::GetTimers(kodi::addon::PVRTimersResultSet& results)
 {
+  if (!m_session->IsConnected()) {
+    return PVR_ERROR_SERVER_ERROR;
+  }
+
   int totals = -1;
   int sum = 0;
   string type = "planned";
@@ -782,7 +613,7 @@ PVR_ERROR TeleBoy::GetTimers(kodi::addon::PVRTimersResultSet& results)
   {
     Document json;
     if (!ApiGet(
-        "/users/" + userId + "/recordings/" + type
+        "/users/" + m_session->GetUserId() + "/recordings/" + type
             + "?desc=1&expand=flags,logos&limit=100&skip=" + to_string(sum) + "&sort=date", json, 10))
     {
       kodi::Log(ADDON_LOG_ERROR, "Error getting recordings of type %s.",
@@ -828,6 +659,10 @@ PVR_ERROR TeleBoy::GetTimers(kodi::addon::PVRTimersResultSet& results)
 
 PVR_ERROR TeleBoy::AddTimer(const kodi::addon::PVRTimer& timer)
 {
+  if (!m_session->IsConnected()) {
+    return PVR_ERROR_SERVER_ERROR;
+  }
+
   if (timer.GetEPGUid() <= EPG_TAG_INVALID_UID)
   {
     return PVR_ERROR_REJECTED;
@@ -836,7 +671,7 @@ PVR_ERROR TeleBoy::AddTimer(const kodi::addon::PVRTimer& timer)
   string postData = "{\"broadcast\": " + to_string(timer.GetEPGUid())
       + ", \"alternative\": false}";
   Document json;
-  if (!ApiPost("/users/" + userId + "/recordings", postData, json))
+  if (!ApiPost("/users/" + m_session->GetUserId() + "/recordings", postData, json))
   {
     kodi::Log(ADDON_LOG_ERROR, "Error recording program %i.", timer.GetEPGUid());
     return PVR_ERROR_SERVER_ERROR;
@@ -849,8 +684,12 @@ PVR_ERROR TeleBoy::AddTimer(const kodi::addon::PVRTimer& timer)
 
 PVR_ERROR TeleBoy::DeleteTimer(const kodi::addon::PVRTimer& timer, bool forceDelete)
 {
+  if (!m_session->IsConnected()) {
+    return PVR_ERROR_SERVER_ERROR;
+  }
+
   Document doc;
-  if (!ApiDelete("/users/" + userId + "/recordings/" + to_string(timer.GetClientIndex()), doc))
+  if (!ApiDelete("/users/" + m_session->GetUserId() + "/recordings/" + to_string(timer.GetClientIndex()), doc))
   {
     kodi::Log(ADDON_LOG_ERROR, "Error deleting timer %i.", timer.GetClientIndex());
     return PVR_ERROR_SERVER_ERROR;
@@ -871,7 +710,11 @@ void TeleBoy::AddTimerType(std::vector<kodi::addon::PVRTimerType>& types, int id
 
 PVR_ERROR TeleBoy::IsEPGTagPlayable(const kodi::addon::PVREPGTag& tag, bool& isPlayable)
 {
-  if (!isComfortMember && !isPlusMember)
+  if (!m_session->IsConnected()) {
+    return PVR_ERROR_SERVER_ERROR;
+  }
+
+  if (!m_session->GetIsPaidMember())
   {
     isPlayable = false;
     return PVR_ERROR_NO_ERROR;
@@ -879,7 +722,7 @@ PVR_ERROR TeleBoy::IsEPGTagPlayable(const kodi::addon::PVREPGTag& tag, bool& isP
 
   time_t current_time;
   time(&current_time);
-  isPlayable = ((current_time - tag.GetEndTime()) < maxRecallSeconds)
+  isPlayable = ((current_time - tag.GetEndTime()) < m_session->GetMaxRecallSeconds())
       && (tag.GetStartTime() < current_time);
   return PVR_ERROR_NO_ERROR;
 }
@@ -888,30 +731,28 @@ PVR_ERROR TeleBoy::IsEPGTagRecordable(const kodi::addon::PVREPGTag& tag, bool& i
 {
   time_t current_time;
   time(&current_time);
-  isRecordable = ((current_time - tag.GetEndTime()) < maxRecallSeconds);
+  isRecordable = ((current_time - tag.GetEndTime()) < m_session->GetMaxRecallSeconds());
   return PVR_ERROR_NO_ERROR;
 }
 
 PVR_ERROR TeleBoy::GetEPGTagStreamProperties(const kodi::addon::PVREPGTag& tag, std::vector<kodi::addon::PVRStreamProperty>& properties)
 {
+  if (!m_session->IsConnected()) {
+    return PVR_ERROR_SERVER_ERROR;
+  }
+
   PVR_ERROR ret = PVR_ERROR_FAILED;
 
   Document json;
   if (!ApiGet(
-      "/users/" + userId + "/stream/"+ to_string(tag.GetUniqueBroadcastId()) + "?" + GetStreamParameters()
+      "/users/" + m_session->GetUserId() + "/stream/"+ to_string(tag.GetUniqueBroadcastId()) + "?" + GetStreamParameters()
           , json, 0))
   {
     kodi::Log(ADDON_LOG_ERROR, "Could not get URL for epg tag.");
     return ret;
   }
-  string url = GetStringOrEmpty(json["data"]["stream"], "url");
-  url = FollowRedirect(url);
-  if (!url.empty())
-  {
-    SetStreamProperties(properties, url, false);
-    ret = PVR_ERROR_NO_ERROR;
-  }
-  return ret;
+  const Value& stream = json["data"]["stream"];
+  return SetStreamProperties(properties, stream, false);
 }
 
 PVR_ERROR TeleBoy::GetEPGTagEdl(const kodi::addon::PVREPGTag& tag, std::vector<kodi::addon::PVREDLEntry>& edl)
@@ -933,67 +774,8 @@ string TeleBoy::GetStringOrEmpty(const Value& jsonValue, const char* fieldName)
   return jsonValue[fieldName].GetString();
 }
 
-bool TeleBoy::ReadDataJson()
-{
-  if (!kodi::vfs::FileExists(data_file, true))
-  {
-    return true;
-  }
-  std::string jsonString = Utils::ReadFile(data_file);
-  if (jsonString.empty())
-  {
-    kodi::Log(ADDON_LOG_ERROR, "Loading data.json failed.");
-    return false;
-  }
-
-  Document doc;
-  doc.Parse(jsonString.c_str());
-  if (doc.GetParseError())
-  {
-    kodi::Log(ADDON_LOG_ERROR, "Parsing data.json failed.");
-    return false;
-  }
-
-  if (doc.HasMember("cinergy_s"))
-  {
-    cinergySCookies = GetStringOrEmpty(doc, "cinergy_s");
-    kodi::Log(ADDON_LOG_DEBUG, "Loaded cinergy_s: %s..", cinergySCookies.substr(0, 5).c_str());
-  }
-
-  kodi::Log(ADDON_LOG_DEBUG, "Loaded data.json.");
-  return true;
-}
-
-bool TeleBoy::WriteDataJson()
-{
-  kodi::vfs::CFile file;
-  if (!file.OpenFileForWrite(data_file, true))
-  {
-    kodi::Log(ADDON_LOG_ERROR, "Save data.json failed.");
-    return false;
-  }
-
-  Document d;
-  d.SetObject();
-  Document::AllocatorType& allocator = d.GetAllocator();
-
-  if (!cinergySCookies.empty())
-  {
-    Value cinergySValue;
-    cinergySValue.SetString(cinergySCookies.c_str(), cinergySCookies.length(), allocator);
-    d.AddMember("cinergy_s", cinergySValue, allocator);
-  }
-
-  StringBuffer buffer;
-  Writer<StringBuffer> writer(buffer);
-  d.Accept(writer);
-  const char* output = buffer.GetString();
-  file.Write(output, strlen(output));
-  return true;
-}
-
 std::string TeleBoy::GetStreamParameters() {
-  std::string params = enableDolby ? "&dolby=1" : "";
+  std::string params = m_session->GetEnableDolby() ? "&dolby=1" : "";
   params += "&https=1&streamformat=dash";
   return params;
 }
